from earthdiagnostics import Utils, cdftools, TempFile
from earthdiagnostics.basins import Basins
from earthdiagnostics.models import Models
from earthdiagnostics.diagnostic import Diagnostic
from earthdiagnostics.box import Box
from autosubmit.config.log import Log
import numpy as np
import os
import shutil
import netCDF4


class Moc(Diagnostic):
    """
    Compute the MOC for oceanic basins
    Created in March 2012
    Author : vguemas@ic3.cat

    # :param input_file:  input grid_V file namez
    # :type input_file: str
    # :param output_file: output file name (=> 2D, depth-y)
    # :param output_file: str
    # :return:
    """

    def __init__(self, data_manager, startdate, member, chunk):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.required_vars = ['vo']
        self.generated_vars = ['vsftmyz']

    @classmethod
    def generate_jobs(cls, diags, options):
        job_list = list()
        for startdate in diags.startdates:
            for member in diags.members:
                for chunk in range(1, diags.chunks + 1):
                    job_list.append(Moc(diags.data_manager, startdate, member, chunk))
        return job_list

    def compute(self):
        temp = TempFile.get()

        input_file = self.data_manager.get_file('ocean', 'vo', self.startdate, self.member, self.chunk)

        Log.debug('Computing MOC')
        cdftools.run('cdfmoc', input=input_file, output=temp)
        Utils.nco.ncks(input=input_file, output=temp, options='-A -v lev')
        Utils.convert2netcdf4(temp)

        Log.debug('Reformatting variables')
        handler = Utils.openCdf(temp)

        handler.createDimension('basin', 5)
        handler.createVariable('basin', str, 'basin')
        handler.variables['basin'][:] = np.array([Basins.Global.fullname, Basins.Atlantic.fullname,
                                                  Basins.Pacific.fullname, Basins.IndoPacific.fullname,
                                                  Basins.Indian.fullname], dtype=object)
        example = handler.variables['zomsfglo']
        # noinspection PyProtectedMember
        moc = handler.createVariable('vsftmyz', example.datatype,
                                     ('time', 'lev', 'i', 'j', 'basin'),
                                     fill_value=example._FillValue)

        moc.units = example.units
        moc.add_offset = example.add_offset
        moc.scale_factor = example.scale_factor

        moc[:, :, :, :, 0] = handler.variables['zomsfglo'][:]
        moc[:, :, :, :, 1] = handler.variables['zomsfatl'][:]
        moc[:, :, :, :, 2] = handler.variables['zomsfpac'][:]
        moc[:, :, :, :, 3] = handler.variables['zomsfinp'][:]
        moc[:, :, :, :, 4] = handler.variables['zomsfind'][:]

        handler.close()

        Utils.nco.ncks(input=temp, output=temp,
                       options='-O -x -v zomsfglo,zomsfatl,zomsfpac,zomsfinp,zomsfind,zomsfinp0')
        Utils.setminmax(temp, 'vsftmyz')

        self.data_manager.send_file(temp, 'ocean', 'vsftmyz', self.startdate, self.member, self.chunk)


class AreaMoc(Diagnostic):
    """
    Compute the MOC for oceanic basins
    Created in March 2012
    Author : vguemas@ic3.cat

    # :param input_file:  input grid_V file namez
    # :type input_file: str
    # :param output_file: output file name (=> 2D, depth-y)
    # :param output_file: str
    # :return:
    """

    def __init__(self, data_manager, startdate, member, chunk, basin, box):
        Diagnostic.__init__(self, data_manager)
        self.basin = basin
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.required_vars = ['vo']
        self.generated_vars = ['vsftmyz']
        self.box = box

    @classmethod
    def generate_jobs(cls, diags, options):
        num_options = len(options) - 1
        if num_options < 4:
            raise Exception('You must specify the box to use')
        if num_options > 5:
            raise Exception('You must specify between 4 and 5 parameters for area moc diagnostic')
        box = Box()
        box.min_lat = int(options[1])
        box.max_lat = int(options[2])
        box.min_depth = int(options[3])
        box.max_depth = int(options[4])
        if num_options > 4:
            basin = Basins.parse(options[5])
        else:
            basin = Basins.Global

        job_list = list()
        for startdate in diags.startdates:
            for member in diags.members:
                for chunk in range(1, diags.chunks + 1):
                    job_list.append(AreaMoc(diags.data_manager, startdate, member, chunk, basin, box))
        return job_list

    def compute(self):
        """
            Compute an Atlantic MOC index by averaging the meridional overturning
            in a latitude band between 1km and 2km
            or any other index averaging the meridional overturning in
            a given basin and a given domain

            Created in March 2012         Author : vguemas@ic3.cat

            :param input_file: input moc file name
            :param lat_min: latitude min
            :param lat_max: latitude max
            :param output_file: output file name      ( => index )
            :param depth_min: depth min
            :param depth_max: depth max
            :param basin: basin
            :return:
            """
        nco = Utils.nco
        cdo = Utils.cdo
        temp2 = TempFile.get()

        temp = self.data_manager.get_file('ocean', 'vsftmyz', self.startdate, self.member, self.chunk)

        handler = Utils.openCdf(temp)
        if 'i' in handler.dimensions:
            handler.close()
            nco.ncwa(input=temp, output=temp, options='-O -a i')
            handler = Utils.openCdf(temp)

        basin_index = np.where(handler.variables['basin'][:] == self.basin.fullname)
        lat_values = handler.variables['lat'][:]
        lat_type = handler.variables['lat'].dtype
        lat_units = handler.variables['lat'].units
        lat_long_name = handler.variables['lat'].long_name

        handler.close()

        if len(basin_index) == 0:
            raise Exception('Basin {0} not defined in file')
        basin_index = basin_index[0][0]
        nco.ncks(input=temp, output=temp, options='-O -d basin,{0}'.format(basin_index))
        # To remove basin dimension
        nco.ncwa(input=temp, output=temp, options='-O -a basin')

        nco.ncks(input=temp, output=temp, options='-O -v vsftmyz,time,lev')

        handler = Utils.openCdf(temp)
        handler.renameDimension('j', 'lat')
        lat_variable = handler.createVariable('lat', lat_type, 'lat')
        lat_variable[:] = lat_values[:]
        lat_variable.units = lat_units
        lat_variable.long_name = lat_long_name

        handler.close()

        nco.ncks(input=temp, output=temp2,
                 options='-O -d lev,{0:.1f},{1:.1f} -d lat,{2:.1f},{3:.1f}'.format(self.box.min_depth, self.box.max_depth,
                                                                                   self.box.min_lat, self.box.max_lat))

        cdo.vertmean(input=temp2, output=temp)
        nco.ncap2(input=temp, output=temp, options='-O -s "coslat[lat]=cos(lat[lat]*3.141592657/180.0)"')
        nco.ncwa(input=temp, output=temp, options='-w coslat -a lat')
        nco.ncks(input=temp, output=temp, options='-O -v vsftmyz,time')
        self.data_manager.send_file(temp, 'ocean', 'vsftmyz', self.startdate, self.member, self.chunk, box=self.box)


class MaxMoc(Diagnostic):

    def __init__(self, data_manager, startdate, member, year, basin, box):
        Diagnostic.__init__(self, data_manager)
        self.basin = basin
        self.startdate = startdate
        self.member = member
        self.year = year
        self.required_vars = ['vo']
        self.generated_vars = ['vsftmyz']
        self.box = box

    @classmethod
    def generate_jobs(cls, diags, options):
        num_options = len(options) - 1
        if num_options < 4:
            raise Exception('You must specify the box to use')
        if num_options > 5:
            raise Exception('You must specify between 4 and 5 parameters for area moc diagnostic')
        box = Box()
        box.min_lat = int(options[1])
        box.max_lat = int(options[2])
        box.min_depth = int(options[3])
        box.max_depth = int(options[4])
        if num_options > 4:
            basin = Basins.parse(options[5])
        else:
            basin = Basins.Global

        job_list = list()
        for startdate in diags.startdates:
            for member in diags.members:
                years = diags.data_manager.get_full_years(startdate)
                if len(years) == 0:
                    Log.user_warning('No complete years are available with the given configuration. '
                                     'MaxMoc can not be computed')
                for year in years:
                    job_list.append(MaxMoc(diags.data_manager, startdate, member, year, basin, box))
        return job_list

    def compute(self):
        """
            Compute an Atlantic MOC index by finding the maximum  of the annual
            mean meridional overturning in a latitude / depth region

            Created in March 2012         Author : vguemas@ic3.cat

            :param basin: basin
            :param input_file: input moc file name
            :param lat_min: latitude min
            :param lat_max: latitude max
            :param output_file: output file name
            :param depth_min: depth mean
            :param depth_max: depth max
            :return:
            """
        nco = Utils.nco

        temp = self.data_manager.get_year('ocean', 'vsftmyz', self.startdate, self.member, self.year)

        handler = Utils.openCdf(temp)
        if 'i' in handler.dimensions:
            handler.close()
            nco.ncwa(input=temp, output=temp, options='-O -a i')
        else:
            handler.close()
        handler = Utils.openCdf(temp)
        basin_index = np.where(handler.variables['basin'][:] == self.basin.fullname)
        if len(basin_index) == 0:
            raise Exception("Basin {1} is not defined in {0}", temp, self.basin.fullname)
        basin_index = basin_index[0][0]

        lev = handler.variables['lev'][:]
        lat = handler.variables['lat'][:]

        if self.box.min_lat == self.box.max_lat:
            lat_inds = ((np.abs(lat-self.box.min_lat)).argmin(),)
        else:
            lat_inds = np.where((lat > self.box.min_lat) & (lat < self.box.max_lat))[0]

        if self.box.min_depth == self.box.max_depth:
            lev_inds = ((np.abs(lev - self.box.min_depth)).argmin(),)
        else:
            lev_inds = np.where((lev > self.box.min_depth) & (lev < self.box.max_depth))[0]

        Log.info('Computing year {0}', str(self.year))
        moc = handler.variables['vsftmyz'][:, lev_inds, lat_inds, basin_index]
        handler.close()
        os.remove(temp)

        moc = np.mean(moc, 0)

        maximum = np.amax(moc)
        max_index = np.unravel_index(np.argmax(moc), moc.shape)
        max_lev = lev[lev_inds[max_index[0]]]
        max_lat = lat[lat_inds[max_index[1]]]

        minimum = np.amin(moc)
        minimum_index = np.unravel_index(np.argmin(moc), moc.shape)
        min_lev = lev[lev_inds[minimum_index[0]]]
        min_lat = lat[lat_inds[minimum_index[1]]]

        Log.info('Maximum {0} Sv, latitude {1} depth {2} m', maximum, max_lat, max_lev)
        Log.info('Minimum {0} Sv, latitude {1} depth {2} m', minimum, min_lat, min_lev)

        handler = self._create_output_file(temp)
        var = handler.createVariable('vsftmyzmax', float, ('time',))
        var.long_name = 'Maximum_Overturing'
        var.units = 'Sverdrup'
        var.valid_min = -1000.
        var.valid_max = 1000.
        var[0] = maximum
        handler.close()
        self.data_manager.send_file(temp, 'ocean', 'vsftmyzmax', self.startdate, self.member, box=self.box,
                                    frequency='yr', year=self.year)

        handler = self._create_output_file(temp)
        var = handler.createVariable('vsftmyzmaxlat', float, ('time',))
        var.long_name = 'Latitude_of_Maximum_Overturing'
        var.units = 'Degrees'
        var.valid_min = -90.
        var.valid_max = 90.
        var[0] = max_lat
        handler.close()
        self.data_manager.send_file(temp, 'ocean', 'vsftmyzmax', self.startdate, self.member, box=self.box,
                                    frequency='yr', year=self.year)

        handler = self._create_output_file(temp)
        var = handler.createVariable('vsftmyzmaxlev', float, ('time',))
        var.long_name = 'Depth_of_Maximum_Overturing'
        var.units = 'Meters'
        var.valid_min = 0.
        var.valid_max = 10000.
        var[0] = max_lev
        handler.close()
        self.data_manager.send_file(temp, 'ocean', 'vsftmyzmax', self.startdate, self.member, box=self.box,
                                    frequency='yr', year=self.year)

        handler = self._create_output_file(temp)
        var = handler.createVariable('vsftmyzmin', float, ('time',))
        var.long_name = 'Minimum_Overtuning'
        var.units = 'Sverdrup'
        var.valid_min = -1000.
        var.valid_max = 1000.
        var[0] = minimum
        handler.close()
        self.data_manager.send_file(temp, 'ocean', 'vsftmyzmax', self.startdate, self.member, box=self.box,
                                    frequency='yr', year=self.year)

        handler = self._create_output_file(temp)
        var = handler.createVariable('vsftmyzminlat', float, ('time',))
        var.long_name = 'Latitude_of_Minimum_Overtuning'
        var.units = 'Degrees'
        var.valid_min = -90.
        var.valid_max = 90.
        var[0] = min_lat
        handler.close()
        self.data_manager.send_file(temp, 'ocean', 'vsftmyzmax', self.startdate, self.member, box=self.box,
                                    frequency='yr', year=self.year)

        handler = self._create_output_file(temp)
        var = handler.createVariable('vsftmyzminlev', float, ('time',))
        var.long_name = 'Depth_of_Minimum_Overtuning'
        var.units = 'Meters'
        var.valid_min = 0.
        var.valid_max = 10000.
        var[0] = min_lev
        handler.close()
        self.data_manager.send_file(temp, 'ocean', 'vsftmyzmax', self.startdate, self.member, box=self.box,
                                    frequency='yr', year=self.year)

    def _create_output_file(self, temp):
        handler = netCDF4.Dataset(temp, 'w')
        handler.createDimension('time')

        time = handler.createVariable('time', 'i2', ('time',))
        time.calendar = 'gregorian'
        time.units = 'days since January 1, {0}'.format(self.year)
        return handler


class Psi(Diagnostic):
    def __init__(self, data_manager, startdate, member, chunk):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.required_vars = ['vo', 'uo']
        self.generated_vars = ['vsftbarot']

    @classmethod
    def generate_jobs(cls, diags, options):
        job_list = list()
        for startdate in diags.startdates:
            for member in diags.members:
                for chunk in range(1, diags.chunks + 1):
                    job_list.append(Psi(diags.data_manager, startdate, member, chunk))
        return job_list

    def compute(self):
        """
        Compute the barotropic stream function

        Created in March 2012         Author : vguemas@ic3.cat
        z
        :param input_file_u: input grid_U file name
        :rtype input_file_U: str
        :param input_file_v: input grid_V file name
        :rtype input_file_V: str
        :param output_file: output file name without nc extension  (=> 2D x-y)
        :rtype output_file: str
        """
        temp = TempFile.get()
        input_file_u = self.data_manager.get_file('ocean', 'uo', self.startdate, self.member, self.chunk)
        input_file_v = self.data_manager.get_file('ocean', 'vo', self.startdate, self.member, self.chunk)
        cdftools.run('cdfpsi', input=[input_file_u, input_file_v], output=temp, options='-mean -mask')
        Utils.rename_variable(temp, 'sobarstf', 'vsftbarot')
        Utils.setminmax(temp, 'vsftbarot')
        self.data_manager.send_file(temp, 'ocean', 'vsftbarot', self.startdate, self.member, self.chunk)


class Gyres(Diagnostic):
    def __init__(self, data_manager, startdate, member, chunk, nemo_version):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.nemo_version = nemo_version
        self.required_vars = ['vsftbarot']
        self.generated_vars = ['gyres']

    @classmethod
    def generate_jobs(cls, diags, options):
        job_list = list()
        for startdate in diags.startdates:
            for member in diags.members:
                for chunk in range(1, diags.chunks + 1):
                    job_list.append(Gyres(diags.data_manager, startdate, member, chunk, diags.nemo_version))
        return job_list

    def compute(self):
        """
            Compute the intensity of the subtropical and subpolar gyres

            Created in October 2013         Author : vguemas@ic3.cat

            :param input_psi: input psi file name
            :param input_grid: input grid
            :param output_file: output file name   ( => index )
            :return:
        """

        if self.nemo_version in [Models.ECEARTH_2_3_O1L42, Models.ECEARTH_3_0_O1L46,
                                 Models.NEMO_3_2_O1L42, Models.NEMO_3_3_O1L46,
                                 Models.NEMOVAR_O1L42]:

            subpolNAtl = [230, 275, 215, 245]
            subpolNPac = [70, 145, 195, 235]
            subtropNPac = [45, 175, 165, 220]
            subtropNAtl = [195, 275, 175, 225]
            subtropSPac = [70, 205, 120, 145]
            subtropSAtl = [235, 300, 120, 145]
            subtropInd = [320, 30, 110, 180]
            ACC = [1, 361, 1, 65]

        elif self in [Models.ECEARTH_3_0_O25L46, Models.ECEARTH_3_0_O25L75,
                      Models.GLORYS2_V1_O25L75]:
            raise Exception("Option gyres not available yet for {0}".format(self.nemo_version))
        else:
            raise Exception("Input grid {0} not recognized".format(self.nemo_version))

        output = TempFile.get()
        vsftbarot_file = self.data_manager.get_file('ocean', 'vsftbarot', self.startdate, self.member, self.chunk)

        handler_original = Utils.openCdf(vsftbarot_file)
        self.var_vsftbarot = handler_original.variables['vsftbarot']
        handler = Utils.openCdf(output, 'w')
        handler.createDimension('time')
        handler.createDimension('region', 8)
        Utils.copy_variable(handler_original, handler, 'time')
        var_region = handler.createVariable('region', str, 'region')

        var_gyre = handler.createVariable('gyre', 'f', ('time', 'region'))
        var_gyre.short_name = 'gyre'
        var_gyre.long_name = 'gyre'
        var_gyre.units = 'm^3/s'

        var_region[0] = 'subpolNAtl'
        var_gyre[:, 0] = self._gyre(subpolNAtl, True)
        Log.info('subpolNAtl: {0}', var_gyre[:, 0])

        var_region[1] = 'subpolNPac'
        var_gyre[:, 1] = self._gyre(subpolNPac, True)
        Log.info('subpolNPac: {0}', var_gyre[:, 1])

        var_region[2] = 'subtropNPac'
        var_gyre[:, 2] = self._gyre(subtropNPac)
        Log.info('subtropNPac: {0}', var_gyre[:, 2])

        var_region[3] = 'subtropSPac'
        var_gyre[:, 3] = self._gyre(subtropSPac)
        Log.info('subtropSPac: {0}', var_gyre[:, 3])

        var_region[4] = 'subtropNAtl'
        var_gyre[:, 4] = self._gyre(subtropNAtl)
        Log.info('subtropNAtl: {0}', var_gyre[:, 4])

        var_region[5] = 'subtropSAtl'
        var_gyre[:, 5] = self._gyre(subtropSAtl)
        Log.info('subtropSAtl: {0}', var_gyre[:, 5])

        var_region[6] = 'subtropInd'
        var_gyre[:, 6] = self._gyre(subtropInd)
        Log.info('subtropInd: {0}', var_gyre[:, 6])

        var_region[7] = 'ACC'
        var_gyre[:, 7] = self._gyre(ACC)
        Log.info('ACC: {0}', var_gyre[:, 7])

        handler.close()
        handler_original.close()
        self.data_manager.send_file(output, 'ocean', 'gyre', self.startdate, self.member, self.chunk)
        Log.info('Finished gyres for startdate {0}, member {1}, chunk {2}', self.startdate, self.member, self.chunk)

    def _gyre(self, site, invert=False):
        """
        Compute the intensity of a given gyre

        Created in October 2013         Author : vguemas@ic3.cat

        :param input_file: input oce file name containing vsftbarot
        :param site: site to calculate convection on
        :param output_file: output file name (=> index)
        :param invert: if True, multiplies result by -1
        :return:
        """

        if invert:
            return np.min(self._extract_section(site), (1, 2)) * -1
        else:
            return np.max(self._extract_section(site), (1, 2))

    def _extract_section(self, site):
        if site[2] <= site[3]:
            if site[0] <= site[1]:
                return self.var_vsftbarot[:, site[2] - 1:site[3] - 1, site[0] - 1:site[1] - 1]
            else:
                return np.concatenate((self.var_vsftbarot[:, site[2] - 1:site[3] - 1, site[0]-1:],
                                      self.var_vsftbarot[:, site[2] - 1:site[3] - 1, :site[1] - 1]), axis=2)

        else:
            if site[0] <= site[1]:
                return np.concatenate((self.var_vsftbarot[:, site[2] - 1:, site[0] - 1: site[1] - 1],
                                      self.var_vsftbarot[:, :site[3] - 1, site[0] - 1: site[1] - 1]), axis=1)
            else:
                temp = np.concatenate((self.var_vsftbarot[:, site[2] - 1:, :],
                                      self.var_vsftbarot[:, :site[3] - 1, :]), axis=1)
                return np.concatenate((temp[:, :, site[0] - 1:],
                                      temp[:, :, :site[1] - 1]), axis=2)



class ConvectionSites(Diagnostic):

    def __init__(self, data_manager, startdate, member, chunk, nemo_version):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.nemo_version = nemo_version
        self.required_vars = ['vsftbarot']
        self.generated_vars = ['gyres']

    @classmethod
    def generate_jobs(cls, diags, options):
        job_list = list()
        for startdate in diags.startdates:
            for member in diags.members:
                for chunk in range(1, diags.chunks + 1):
                    job_list.append(ConvectionSites(diags.data_manager, startdate, member, chunk, diags.nemo_version))
        return job_list

    def compute(self):
        """
        Compute the intensity of convection in the four main convection sites

        Created in October 2013         Author : vguemas@ic3.cat

        :param input_file: input oce file name containing mlotst
        :param input_grid: input grid
        :param output_file: output file name (=> index)
        :return:
        """
        if self.nemo_version  in [Models.ECEARTH_2_3_O1L42, Models.ECEARTH_3_0_O1L46,
                                  Models.NEMO_3_2_O1L42, Models.NEMO_3_3_O1L46,
                                  Models.NEMOVAR_O1L42]:
            labrador = [225, 245, 215, 255]
            irminger = [245, 290, 215, 245]
            gin = [260, 310, 245, 291]
            wedell = [225, 280, 1, 50]

        elif self.nemo_version in [Models.ECEARTH_3_0_O25L46, Models.ECEARTH_3_0_O25L75,
                                   Models.GLORYS2_V1_O25L75]:
            raise Exception("Option convection not available yet for {0}".format(self.nemo_version))
        else:
            raise Exception("Input grid {0} not recognized".format(self.nemo_version))

        mlotst_file = self.data_manager.get_file('ocean', 'mlotst', self.startdate, self.member, self.chunk)
        output = TempFile.get()

        self.mlotst_handler = Utils.openCdf(mlotst_file)
        handler = Utils.openCdf(output, 'w')
        handler.createDimension('time')
        handler.createDimension('region', 4)
        Utils.copy_variable(self.mlotst_handler, handler, 'time')
        var_region = handler.createVariable('region', str, 'region')
        var_gyre = handler.createVariable('site', 'f', ('time', 'region'))
        var_gyre.short_name = 'site'
        var_gyre.long_name = 'convection sites'
        var_gyre.units = 'm^3/s'

        var_region[0] = 'labrador'
        var_gyre[:, 0] = self._convection_site(labrador)

        var_region[1] = 'irminger'
        var_gyre[:, 1] = self._convection_site(irminger)

        var_region[2] = 'gin'
        var_gyre[:, 2] = self._convection_site(gin)

        var_region[3] = 'wedell'
        var_gyre[:, 3] = self._convection_site(wedell)

        self.mlotst_handler.close()
        handler.close()

        self.data_manager.send_file(output, 'ocean', 'site', self.startdate, self.member, self.chunk)
        Log.info('Finished convection sites for startdate {0}, member {1}, chunk {2}',
                 self.startdate, self.member, self.chunk)

    def _convection_site(self, site):
        return np.max(self.mlotst_handler.variables['mlotst'][:, site[2]-1:site[3]-1, site[0]-1:site[1]-1], (1, 2))

